package fust2013.aop.generics.aspects;

import java.util.LinkedList;
import java.util.List;
import java.util.WeakHashMap;

public abstract aspect ObserverProtocol<Subject,Observer> {
	private WeakHashMap<Subject, List<Observer>> perSubjectObservers;
	
	public List<Observer> getObservers(Subject s) {
		if (perSubjectObservers == null) {
			System.out.println("creating perSubjectObservers in this:"+this.toString());
			perSubjectObservers = new WeakHashMap<>();
		}
		List<Observer> observers = perSubjectObservers.get(s);
		if (observers == null) {
			observers = new LinkedList<>();
			perSubjectObservers.put(s, observers);
		}
		return observers;
	}

	public void addObserver(Subject s, Observer o) {
		System.out.println("adding "+o+" for subject "+s);
		getObservers(s).add(o);
	}
	
	public void removeObserver(Subject s, Observer o) {
		System.out.println("removing "+o+" from subject "+s);
		getObservers(s).remove(o);
	}

	abstract protected pointcut subjectChange(Subject s);

	abstract protected void updateObserver(Subject s, Observer o);
	
	after(Subject s) : subjectChange(s) {
		for (Observer o : getObservers(s))
			updateObserver(s, o);
	}
}